from datetime import datetime
from unittest.mock import patch

from jdatetime import datetime as jdatetime

# noinspection PyProtectedMember
from tsetmc import Instrument, _core, get_market_watch_init, \
    get_closing_price_all, get_client_type_all, get_key_stats


def patch_get_content(name):
    with open(f'{__file__}/../data/{name}', 'rb') as f:
        text = f.read()
    return patch.object(_core, 'get_content', lambda _: text)


@patch_get_content('fmelli.html')
def test_get_page_info():
    d = Instrument(1).get_page_info()
    assert d == {
        'tmax': 19820.0,
        'tmin': 21900.0,
        'bvol': 5479452,
        'eps': 1639,
        'free_float': 33,
        'l30': 'ملی\u200c صنایع\u200c مس\u200c ایران\u200c',  # todo
        'sector_name': 'فلزات اساسی',
        'market': 'بازار اول (تابلوی اصلی) بورس',
        'month_average_volume': 98322903,
        'l18': 'فملی',
        'sector_pe': 15.4,
        'z': 101400000000,
        'week_max': 22000.0,
        'week_min': 20010.0,
        'year_max': 39810.0,
        'year_min': 5181.0}


@patch_get_content('dey.html')
def test_get_page_info_no_free_float():
    d = Instrument(1).get_page_info()
    assert d == {
        'tmax': 43722.0,
        'tmin': 48324.0,
        'bvol': 2231446,
        'eps': 4326,
        'free_float': None,
        'l30': 'بانک دی',
        'sector_name': 'بانکها و موسسات اعتباری',
        'market': 'بازار دوم فرابورس',
        'month_average_volume': 25836324,
        'l18': 'دی',
        'sector_pe': 16.69,
        'z': 6400000000,
        'week_max': 49934.0,
        'week_min': 43722.0,
        'year_max': 83000.0,
        'year_min': 2674.0}


@patch_get_content('kala.html')
def test_get_page_info_no_eps():
    d = Instrument(1).get_page_info()
    assert d == {
        'tmax': 37970.0,
        'tmin': 41950.0,
        'bvol': 1306165,
        'eps': None,
        'free_float': 100,
        'l30': 'بورس کالای ایران',
        'sector_name': 'فعالیتهای کمکی به نهادهای مالی واسط',
        'market': 'بازار اول (تابلوی فرعی) بورس',
        'month_average_volume': 5720005,
        'l18': 'کالا',
        'sector_pe': 26.22,
        'z': 2500000000,
        'week_max': 46870.0,
        'week_min': 37970.0,
        'year_max': 105352.0,
        'year_min': 12376.0}


@patch_get_content('khgostar.html')
def test_get_page_info_no_sector_pe():
    d = Instrument(1).get_page_info()
    assert d == {
        'tmax': 3670.0,
        'tmin': 4050.0,
        'bvol': 15842055,
        'eps': 450,
        'free_float': 52,
        'l30': 'گسترش\u200cسرمایه\u200cگذاری\u200cایران\u200cخودرو',
        'sector_name': 'خودرو و ساخت قطعات',
        'market': 'بازار اول (تابلوی اصلی) بورس',
        'month_average_volume': 212159873,
        'l18': 'خگستر',
        'sector_pe': None,
        'z': 39605137000,
        'week_max': 4250.0,
        'week_min': 3670.0,
        'year_max': 21750.0,
        'year_min': 2616.0}


@patch_get_content('dara_yekom.txt')
def test_dara1_instant():
    assert Instrument(1).get_info() == {
        'pc': 181240,
        'pmax': 176300,
        'pmin': 185000,
        'pf': 181600,
        'last_info_datetime': datetime(2020, 11, 11, 12, 29, 59),
        'pl': 183250,
        'nav': 202821,
        'nav_datetime': jdatetime(1399, 8, 21, 11, 56),
        'tno': 62630,
        'timestamp': '12:29:59',
        'tval': 6085868910210,
        'tvol': 33579648,
        'py': 181110}


@patch_get_content('asam.txt')
def test_asam_instant():
    assert Instrument(1).get_info() == {
        'pc': 94140,
        'pmax': 92001,
        'pmin': 96000,
        'last_info_datetime': datetime(2020, 11, 11, 12, 28, 17),
        'pl': 95890,
        'nav': 95630,
        'nav_datetime': jdatetime(1399, 8, 21, 15, 13, 43),
        'tno': 27,
        'pf': 94440,
        'timestamp': '12:28:17',
        'tval': 753116350,
        'tvol': 8000,
        'py': 93414
    }


@patch_get_content('negin.txt')
def test_negin_instant():
    assert Instrument(1).get_info() == {
        'pc': 50110,
        'pmax': 50000,
        'pmin': 50700,
        'last_info_datetime': datetime(2020, 11, 11, 12, 29, 37),
        'pl': 50000,
        'nav': 12190,
        'nav_datetime': jdatetime(1398, 12, 29, 16, 0),
        'tno': 29,
        'pf': 50010,
        'timestamp': '12:29:37',
        'tval': 1408671580,
        'tvol': 28109,
        'py': 50010}


@patch_get_content('fmelli.txt')
def test_fmelli_instant():
    assert Instrument(1).get_info() == {
        'pc': 19890,
        'pmax': 19200,
        'pmin': 20320,
        'last_info_datetime': datetime(2020, 11, 11, 17, 29, 53),
        'pl': 20320,
        'nav': None,
        'nav_datetime': None,
        'tno': 26572,
        'pf': 19400,
        'timestamp': '17:29:53',
        'tval': 1318025925250,
        'tvol': 66266936,
        'py': 19360}


@patch_get_content('MarketWatchInit.aspx')
def test_get_market_watch_init():
    df = get_market_watch_init()['dataframe']
    assert df.dtypes.to_dict() == {
        'heven': 'int64', 'pf': 'int64', 'pc': 'int64',
        'pl': 'int64', 'tno': 'int64', 'tvol': 'int64',
        'tval': 'int64', 'pmin': 'int64', 'pmax': 'int64',
        'py': 'int64', 'eps': 'float64', 'bvol': 'int64',
        'visitcount': 'int64', 'flow': 'int64',
        'cs': 'int64', 'tmax': 'Int64', 'tmin': 'Int64',
        'z': 'int64', 'yval': 'int64', 'pd1': 'Int64',
        'po1': 'Int64', 'qd1': 'Int64', 'qo1': 'Int64',
        'zd1': 'Int64', 'zo1': 'Int64', 'pd2': 'Int64',
        'po2': 'Int64', 'qd2': 'Int64', 'qo2': 'Int64',
        'zd2': 'Int64', 'zo2': 'Int64', 'pd3': 'Int64',
        'po3': 'Int64', 'qd3': 'Int64', 'qo3': 'Int64',
        'zd3': 'Int64', 'zo3': 'Int64'}
    index = df.index
    assert index.names == ['id', 'isin', 'l18', 'l30']
    assert index.dtype == 'O'
    # in pandas 1.3 there will be no need for `.to_frame()`
    assert index.to_frame().dtypes.to_list() == ['int64', 'O', 'O', 'O']


@patch_get_content('ClosingPriceAll.aspx')
def test_get_closing_price_all():
    df = get_closing_price_all()
    assert df.dtypes.to_dict() == {
        'PClosing': 'int64', 'PDrCotVal': 'int64', 
        'ZTotTran': 'int64', 'QTotTran5J': 'int64', 
        'QTotCap': 'int64', 'PriceMin': 'int64', 
        'PriceMax': 'int64', 'PriceYesterday': 'int64', 
        'PriceFirst': 'int64'}
    index = df.index
    assert index.names == ['id', 'n']
    assert index.dtype == 'O'
    # in pandas 1.3 there will be no need for `.to_frame()`
    assert index.to_frame().dtypes.to_list() == ['int64', 'int64']


@patch_get_content('ClientTypeAll.aspx')
def test_get_client_type_all():
    df = get_client_type_all()
    assert all(df.columns == [
        'Buy_CountI', 'Buy_CountN', 'Buy_I_Volume', 'Buy_N_Volume',
        'Sell_CountI', 'Sell_CountN', 'Sell_I_Volume', 'Sell_N_Volume'])
    assert all(dt == 'int64' for dt in df.dtypes)
    assert df.index.name == 'id'


@patch_get_content('InstValue.aspx')
def test_get_key_stats():
    df = get_key_stats()
    assert all(df.columns.str.startswith('is'))
    assert df.index.name == 'id'


VSKHOOZ = {
    'last_info_datetime': datetime(2021, 1, 20, 6, 42, 18),
    'nav': None,
    'nav_datetime': None,
    'pc': 67818,
    'pf': 0,
    'pl': 67810,
    'pmax': 0,
    'pmin': 0,
    'py': 67818,
    'timestamp': '06:42:18',
    'tno': 0,
    'tval': 0,
    'tvol': 0}


@patch_get_content('vskhooz_short_response.txt')
def test_vskhooz_short():
    assert Instrument(1).get_info() == VSKHOOZ


@patch_get_content('vskhooz_long_response.txt')
def test_vskhooz_long():
    assert Instrument(1).get_info() == VSKHOOZ


@patch_get_content('fmelli_trade_history_top2.txt')
def test_get_trade_history():
    df = Instrument(1).get_trade_history(2)
    assert df.to_dict() == {
        'pmax': {20210120: 10400.0, 20210119: 10380.0}
        , 'pmin': {20210120: 10120.0, 20210119: 9400.0}
        , 'pc': {20210120: 10380.0, 20210119: 9910.0}
        , 'pl': {20210120: 10400.0, 20210119: 10290.0}
        , 'pf': {20210120: 10350.0, 20210119: 9400.0}
        , 'py': {20210120: 9910.0, 20210119: 9890.0}
        , 'tval': {20210120: 498484813880.0, 20210119: 2649416188110.0}
        , 'tvol': {20210120: 48013394, 20210119: 267389256}
        , 'tno': {20210120: 7284, 20210119: 36765}}
    index = df.index
    assert index.name == 'date'
    assert all(index.values == [20210120, 20210119])



from re import compile as rc
from datetime import datetime
from io import BytesIO, StringIO

from requests import Session
from jdatetime import datetime as jdatetime
from pandas import read_csv, to_numeric, DataFrame


strptime = datetime.strptime
jstrptime = jdatetime.strptime
GET = Session().get


FARSI_NORM = ''.maketrans('يك', 'یک')
F = r'([\d\.]+)'
SECTOR_PE_SEARCH = rc(rf"SectorPE='{F}'").search
TITLE_SEARCH = rc(r"Title='(.*?) \((.*?)\) \- ([^']*)'").search
FREE_FLOAT_SEARCH = rc(r"KAjCapValCpsIdx='([\d\.]+)'").search
GROUP_NAME_SEARCH = rc(r"LSecVal='(.*?)'").search
BASE_VOLUME_SEARCH = rc(r"BaseVol=(\d+)").search
EPS_SEARCH = rc(r"EstimatedEPS='(\d+)'").search
SHARES_SEARCH = rc(r'ZTitad=(\d+)').search
ALLOWED_MIN_MAX_SEARCH = rc(rf"PSGelStaMax='{F}',PSGelStaMin='{F}").search
WEAK_YEAR_MIN_MAX_SEARCH = rc(
    rf"MinWeek='{F}',MaxWeek='{F}',MinYear='{F}',MaxYear='{F}'").search
MONTH_AVG_VOL_SEARCH = rc(r"QTotTran5JAvg='(\d+)'").search
FIRST_NUMBER_SEARCH = rc(r'\d+').search
INSTANT_MATCH = rc(
    r'(?P<timestamp>\d\d:\d\d:\d\d),'
    r'..,'
    r'(?P<pl>\d+),'
    r'(?P<pc>\d+),'
    r'(?P<pf>\d+),'
    r'(?P<py>\d+),'
    r'(?P<pmin>\d+),'
    r'(?P<pmax>\d+),'
    r'(?P<tno>\d+),'
    r'(?P<tvol>\d+),'
    r'(?P<tval>\d+),'
    r'1,'
    r'(?P<last_info_datetime>\d+,\d+)'
    r'(,(?P<nav_datetime>[\d\/: ]+),(?P<nav>\d+))?'
).match
INSTANT_INTS = {
    'pl', 'pc', 'py', 'pf', 'py', 'pmin', 'pmax', 'tno', 'tvol', 'tval'}


def get_content(url) -> bytes:
    return GET(url).content


def fa_norm_text(url) -> str:
    # replace Arabic [ي ك] with Persian [ی ک]
    return get_content(url).decode().translate(FARSI_NORM)


class Instrument:
    # warning/todo:
    # get_page_info and get_inst_info are not tested widely and fail sometimes.

    def __init__(self, id: int):
        self.id = id

    def get_page_info(self) -> dict:
        """Return the static info of the symbol's page.

        For the meaning of keys see:
            https://cdn.tsetmc.com/Site.aspx?ParTree=151713
        """
        text = fa_norm_text(f'http://tsetmc.com/Loader.aspx?ParTree=151311&i={self.id}')
        t_min_max = ALLOWED_MIN_MAX_SEARCH(text)
        wy_min_max = WEAK_YEAR_MIN_MAX_SEARCH(text)
        title_match = TITLE_SEARCH(text)
        free_float_match = FREE_FLOAT_SEARCH(text)
        eps_match = EPS_SEARCH(text)
        sector_pe_match = SECTOR_PE_SEARCH(text)
        return {
            'tmax': float(t_min_max[2]),
            'tmin': float(t_min_max[1]),
            'bvol': int(BASE_VOLUME_SEARCH(text)[1]),
            'eps': int(eps_match[1]) if eps_match is not None else None,
            'free_float': int(
                free_float_match[1]) if free_float_match is not None else None,
            'l30': title_match[1],
            'sector_name': GROUP_NAME_SEARCH(text)[1],
            'market': title_match[3],
            'month_average_volume': int(MONTH_AVG_VOL_SEARCH(text)[1]),
            'l18': title_match[2],
            'sector_pe': float(
                sector_pe_match[1]) if sector_pe_match is not None else None,
            'z': int(SHARES_SEARCH(text)[1]),
            'week_max': float(wy_min_max[2]),
            'week_min': float(wy_min_max[1]),
            'year_max': float(wy_min_max[4]),
            'year_min': float(wy_min_max[3]),
        }

    def get_info(self) -> dict:
        # apparently, http://www.tsetmc.com/tsev2/data/instinfodata.aspx?i=...
        # and http://www.tsetmc.com/tsev2/data/instinfofast.aspx?i=...
        # return the same response.
        text = fa_norm_text(
            f'http://www.tsetmc.com/tsev2/data/instinfodata.aspx'
            f'?i={self.id}&c=')
        group_dict = INSTANT_MATCH(text).groupdict()
        for k in INSTANT_INTS:
            # noinspection PyTypeChecker
            group_dict[k] = int(group_dict[k])
        # noinspection PyTypeChecker
        group_dict['last_info_datetime'] = strptime(
            group_dict['last_info_datetime'], '%Y%m%d,%H%M%S')
        if (nav_datetime := group_dict.get('nav_datetime')) is not None:
            # noinspection PyTypeChecker
            group_dict['nav'] = int(group_dict['nav'])
            group_dict['nav_datetime'] = jstrptime(
                nav_datetime, '%Y/%m/%d %H:%M:%S')
        return group_dict

    def get_trade_history(self, top: int) -> DataFrame:
        content = get_content(
            f'http://members.tsetmc.com/tsev2/data/InstTradeHistory.aspx?i={self.id}&Top={top}')
        df = read_csv(
            BytesIO(content)
            , sep='@'
            , lineterminator=';'
            , names=['date', 'pmax', 'pmin', 'pc', 'pl', 'pf', 'py', 'tval', 'tvol', 'tno']
            , low_memory=False
            , index_col='date'
        )
        return df

    @staticmethod
    def from_search(s: str) -> 'Instrument':
        text = fa_norm_text('http://tsetmc.com/tsev2/data/search.aspx?skey=' + s)
        return Instrument(int(FIRST_NUMBER_SEARCH(text)[0]))


def get_market_watch_init() -> dict:
    """Return the market status which are the info used in creating filters.

    For more information about filters see:
        http://tsetmc.com/Loader.aspx?ParTree=15131F
    For the meaning of column names in the returned DataFrame see:
        https://cdn.tsetmc.com/Site.aspx?ParTree=151713
        For `flow` and `yval` codes see:
            http://cdn.tsetmc.com/Site.aspx?ParTree=1114111118&LnkIdn=83
        For `heven` see:
            http://cdn.tsetmc.com/Site.aspx?ParTree=111411111S&LnkIdn=129
            (it's the time of last transaction in HHMMSS format)
    """
    text = fa_norm_text('http://tsetmc.com/tsev2/data/MarketWatchInit.aspx?h=0&r=0')
    _, _, states, price_rows, _ = text.split('@')
    state_df = read_csv(
        StringIO(states),
        lineterminator=';',
        names=(
            'id', 'isin', 'l18', 'l30', 'heven', 'pf', 'pc', 'pl', 'tno',
            'tvol', 'tval', 'pmin', 'pmax', 'py', 'eps', 'bvol', 'visitcount',
            'flow', 'cs', 'tmax', 'tmin', 'z', 'yval'),
        # unlike int64, Int64 is nullable
        dtype={'tmin': "Int64", 'tmax': "Int64"},
        low_memory=False,
        index_col=['id', 'isin', 'l18', 'l30'])
    price_df = read_csv(
        StringIO(price_rows),
        lineterminator=';',
        names=('id', 'row', 'zo', 'zd', 'pd', 'po', 'qd', 'qo'),
        dtype="Int64",
        low_memory=False)
    # merge multiple rows sharing the same `row` number into one row.
    # a fascinating solution from https://stackoverflow.com/a/53563551/2705757
    price_df.set_index(['id', 'row'], inplace=True)
    price_df = price_df.unstack(fill_value=0).sort_index(1, 1)
    price_df.columns = [f'{c}{i}' for c, i in price_df.columns]
    joined_df = state_df.join(price_df)
    # joined_df.index = to_numeric(joined_df.index, downcast='unsigned')
    return {  # todo, also add other info available in MarketWatchInit.aspx
        'dataframe': joined_df,
    }


def _split_id_rows(content: bytes, id_row_len: int) -> list:
    data = content.split(b';')
    for i, datum in enumerate(data):
        items = datum.split(b',')
        if len(items) == id_row_len:
            id_ = items[0]
        else:
            # noinspection PyUnboundLocalVariable
            items.insert(0, id_)
        # noinspection PyTypeChecker
        data[i] = items
    return data


def get_closing_price_all() -> DataFrame:
    """Return price history dataframe.

    For the meaning of column names refer to
        http://tsetmc.com/Site.aspx?ParTree=151715&LnkIdn=3197
    """
    content = get_content('http://members.tsetmc.com/tsev2/data/ClosingPriceAll.aspx')
    data = _split_id_rows(content, id_row_len=11)
    df = DataFrame(data, columns=(
        'id', 'n', 'PClosing', 'PDrCotVal', 'ZTotTran', 'QTotTran5J',
        'QTotCap', 'PriceMin', 'PriceMax', 'PriceYesterday', 'PriceFirst'))
    # noinspection PyTypeChecker
    df = df.apply(to_numeric)
    df.set_index(['id', 'n'], inplace=True)
    return df


def get_client_type_all() -> DataFrame:
    """Return client types (natural/legal stats) as a DataFrame.

    For the meaning of column names refer to
        https://cdn.tsetmc.com/Site.aspx?ParTree=151715&LnkIdn=3198
    """
    content = get_content('http://www.tsetmc.com/tsev2/data/ClientTypeAll.aspx')
    df = read_csv(
        BytesIO(content), lineterminator=b';', names=(
            'id', 'Buy_CountI', 'Buy_CountN', 'Buy_I_Volume', 'Buy_N_Volume',
            'Sell_CountI', 'Sell_CountN', 'Sell_I_Volume', 'Sell_N_Volume'),
        dtype="int64", index_col='id', low_memory=False)
    return df


def get_key_stats() -> DataFrame:
    """Return key statistics as a DataFrame.

    For the meaning of column names refer to
        http://www.tsetmc.com/Site.aspx?ParTree=151715&LnkIdn=3199 or
        http://cdn.tsetmc.com/Site.aspx?ParTree=151713
    """
    content = get_content('http://www.tsetmc.com/tsev2/data/InstValue.aspx?t=a')
    data = _split_id_rows(content, id_row_len=3)
    df = DataFrame(data, columns=('id', 'n', 'value'))
    # noinspection PyTypeChecker
    df = df.apply(to_numeric)
    df = df.pivot('id', 'n', 'value')
    df.columns = [f'is{c}' for c in df.columns]
    return df
